Index: mcomp.c
===================================================================
RCS file: /usr/local/cvsroot/theora/lib/mcomp.c,v
retrieving revision 1.5
diff -u -r1.5 mcomp.c
--- mcomp.c	23 Sep 2002 08:31:02 -0000	1.5
+++ mcomp.c	8 May 2003 22:35:09 -0000
@@ -19,6 +19,682 @@
 #include <ogg/ogg.h>
 #include "encoder_internal.h"
 
+/*
+ * Summary of functions provided both in C, MMX and MMX-extended
+ * implementations (see HAVE_MMX and HAVE_MMXEXT macros below).
+ * A priori, the C implementation is the reference.
+ */
+/* Does the sum of absolute differences of all coordinates of 2 vectors
+   of 8 bytes values. */
+static inline ogg_uint32_t sum_abs_diff8(unsigned char * NewDataPtr, 
+					 unsigned char  * RefDataPtr);
+/* Does the sum of the absolute differences of all coordinates between
+   1 vector of 8 values, and the average of 2 vectors of 8 values. */
+static inline ogg_uint32_t sum_abs_diffmean8(unsigned char * SrcData, 
+					     unsigned char * RefDataPtr1,
+					     unsigned char * RefDataPtr2);
+/* Computes the differences of 2 vectors of 8 values, and returns the sum
+   of the coordinates of the resulting 8 values vectors as well as the sum
+   of the square of these coordinates. */
+static inline void accsumsqr_diff8(unsigned char *new, unsigned char *ref,
+				   ogg_int32_t *accsum, ogg_int32_t *accsum2);
+/* Computes the difference between 1 vector of 8 values and the average of
+   2 vectors of 8 values, and then returns the sum of the coordinates of the
+   resulting 8 values vectors as well as the sum of the square of these
+   coordinates. */
+static inline void accsumsqr_diffmean8(unsigned char *new,
+				   unsigned char *ref1, unsigned char *ref2,
+				   ogg_int32_t *accsum, ogg_int32_t *accsum2);
+/* Computes the sum and the sum of the square of the coordinates of an
+   8 value vector. */
+static inline void accsumsqr8(unsigned char *diff,
+			      ogg_int32_t *accsum, ogg_int32_t *accsum2);
+/* This function must be called after calling the former functions and before
+   performing any floating point operations (MMX necessitates this, the C
+   variant is void). */
+static inline void finish8(void);
+
+/*
+ * C reference implementation of basic vector operations
+ * (see below for MMX accelerated assembler variants)
+ */
+static inline ogg_uint32_t sum_abs_diff8_c(unsigned char * NewDataPtr, 
+					   unsigned char  * RefDataPtr) {
+  ogg_uint32_t Res; /* Final result */
+  Res = abs( ((int)NewDataPtr[0]) - ((int)RefDataPtr[0]) );
+  Res += abs( ((int)NewDataPtr[1]) - ((int)RefDataPtr[1]) );
+  Res += abs( ((int)NewDataPtr[2]) - ((int)RefDataPtr[2]) );
+  Res += abs( ((int)NewDataPtr[3]) - ((int)RefDataPtr[3]) );
+  Res += abs( ((int)NewDataPtr[4]) - ((int)RefDataPtr[4]) );
+  Res += abs( ((int)NewDataPtr[5]) - ((int)RefDataPtr[5]) );
+  Res += abs( ((int)NewDataPtr[6]) - ((int)RefDataPtr[6]) );
+  Res += abs( ((int)NewDataPtr[7]) - ((int)RefDataPtr[7]) );
+
+  return Res;
+}
+static inline ogg_uint32_t sum_abs_diffmean8_c(unsigned char * SrcData, 
+					       unsigned char * RefDataPtr1,
+					       unsigned char * RefDataPtr2) {
+  ogg_uint32_t Res; /* Final result */
+  Res = abs( ((int)SrcData[0]) - (((int)RefDataPtr1[0] + 
+				   (int)RefDataPtr2[0]) / 2) );
+  Res += abs( ((int)SrcData[1]) - (((int)RefDataPtr1[1] + 
+				    (int)RefDataPtr2[1]) / 2) );
+  Res += abs( ((int)SrcData[2]) - (((int)RefDataPtr1[2] + 
+				    (int)RefDataPtr2[2]) / 2) );
+  Res += abs( ((int)SrcData[3]) - (((int)RefDataPtr1[3] + 
+				    (int)RefDataPtr2[3]) / 2) );
+  Res += abs( ((int)SrcData[4]) - (((int)RefDataPtr1[4] + 
+				    (int)RefDataPtr2[4]) / 2) );
+  Res += abs( ((int)SrcData[5]) - (((int)RefDataPtr1[5] + 
+				    (int)RefDataPtr2[5]) / 2) );
+  Res += abs( ((int)SrcData[6]) - (((int)RefDataPtr1[6] + 
+				    (int)RefDataPtr2[6]) / 2) );
+  Res += abs( ((int)SrcData[7]) - (((int)RefDataPtr1[7] + 
+				    (int)RefDataPtr2[7]) / 2) );
+  return Res;
+}
+
+static inline void accsumsqr_diff8_c(unsigned char *new, unsigned char *ref,
+				     ogg_int32_t *accsum, ogg_int32_t *accsum2) {
+  ogg_int32_t diff, t, t2;
+
+  t = 0;
+  t2 = 0;
+  diff= (int)new[0] - (int)ref[0];
+  t += diff;
+  t2 += diff*diff;
+  diff= (int)new[1] - (int)ref[1];
+  t += diff;
+  t2 += diff*diff;
+  diff= (int)new[2] - (int)ref[2];
+  t += diff;
+  t2 += diff*diff;
+  diff= (int)new[3] - (int)ref[3];
+  t += diff;
+  t2 += diff*diff;
+  diff= (int)new[4] - (int)ref[4];
+  t += diff;
+  t2 += diff*diff;
+  diff= (int)new[5] - (int)ref[5];
+  t += diff;
+  t2 += diff*diff;
+  diff= (int)new[6] - (int)ref[6];
+  t += diff;
+  t2 += diff*diff;
+  diff= (int)new[7] - (int)ref[7];
+  t += diff;
+  t2 += diff*diff;
+  (*accsum) += t;
+  (*accsum2) += t2;
+}
+
+static inline void accsumsqr_diffmean8_c(unsigned char *new,
+					 unsigned char *ref1,
+					 unsigned char *ref2,
+					 ogg_int32_t *accsum,
+					 ogg_int32_t *accsum2) {
+  ogg_int32_t diff, t, t2;
+
+  t = 0;
+  t2 = 0;
+  diff= ((int)new[0]) - (((int)ref1[0] + ref2[0]) / 2);
+  t += diff;
+  t2 += diff*diff;
+  diff= ((int)new[1]) - (((int)ref1[1] + ref2[1]) / 2);
+  t += diff;
+  t2 += diff*diff;
+  diff= ((int)new[2]) - (((int)ref1[2] + ref2[2]) / 2);
+  t += diff;
+  t2 += diff*diff;
+  diff= ((int)new[3]) - (((int)ref1[3] + ref2[3]) / 2);
+  t += diff;
+  t2 += diff*diff;
+  diff= ((int)new[4]) - (((int)ref1[4] + ref2[4]) / 2);
+  t += diff;
+  t2 += diff*diff;
+  diff= ((int)new[5]) - (((int)ref1[5] + ref2[5]) / 2);
+  t += diff;
+  t2 += diff*diff;
+  diff= ((int)new[6]) - (((int)ref1[6] + ref2[6]) / 2);
+  t += diff;
+  t2 += diff*diff;
+  diff= ((int)new[7]) - (((int)ref1[7] + ref2[7]) / 2);
+  t += diff;
+  t2 += diff*diff;
+
+  (*accsum) += t;
+  (*accsum2) += t2;
+}
+
+static inline void accsumsqr8_c(unsigned char *diff,
+				ogg_int32_t *accsum, ogg_int32_t *accsum2) {
+  ogg_int32_t t, t2;
+
+  t = 0;
+  t2 = 0;
+  t += diff[0];
+  t2 += diff[0]*diff[0];
+  t += diff[1];
+  t2 += diff[1]*diff[1];
+  t += diff[2];
+  t2 += diff[2]*diff[2];
+  t += diff[3];
+  t2 += diff[3]*diff[3];
+  t += diff[4];
+  t2 += diff[4]*diff[4];
+  t += diff[5];
+  t2 += diff[5]*diff[5];
+  t += diff[6];
+  t2 += diff[6]*diff[6];
+  t += diff[7];
+  t2 += diff[7]*diff[7];
+
+  (*accsum) += t;
+  (*accsum2) += t2;
+}
+
+static inline void finish8_c(void) {
+  /* Nothing to do */
+}
+
+/*
+ * Use the following #define to trigger conditional compilation
+ *  HAVE_MMX    : provide MMX-based assembler functions
+ *  HAVE_MMXEXT : use MMX-extended functions (more precisely the psadbw
+ *     instruction to compute averages or sum of absolute differences
+ *     for 8 bytes vectors). Usually comes with SSE or 3DNow! or whatever...
+ *  TEST_MMX    : the wrapper functions which selects C or assembly
+ *     variants are changed to provide back-to-back testing of both
+ *     implementations and report differences to stderr if any. Useful
+ *     for checking the assembly... (But kills performance of course as
+ *     both implementations are executed successively.)
+ */
+
+
+/* #define HAVE_MMX */
+/* #define HAVE_MMXEXT */
+/* #define TEST_MMX */
+
+
+#ifdef HAVE_MMX
+
+#ifdef HAVE_MMXEXT
+
+/*
+ * extended MMX (with psadbw and pavgb instruction!) implementation
+ */
+static inline ogg_uint32_t sum_abs_diff8_mmxext(unsigned char * NewDataPtr, 
+						unsigned char * RefDataPtr) {
+  unsigned long long int *new = (unsigned long long int*)NewDataPtr;
+  unsigned long long int *ref = (unsigned long long int*)RefDataPtr;
+  ogg_uint32_t Res; /* Final result */
+  __asm__("\n\t movq %1,%%mm0        \t # load new"
+	  "\n\t psadbw %2,%%mm0      \t # does the operation"
+	  "\n\t movd %%mm0,%0        \t # store result"
+	  : "=m" (Res) /* this is %0, output */
+	  : "m" (*new),      /* this is %1, new */
+	  "m" (*ref)      /* this is %2, ref */
+	  );
+  return Res;
+}
+
+static inline ogg_uint32_t sum_abs_diffmean8_mmxext(unsigned char * SrcData, 
+						    unsigned char * RefDataPtr1,
+						 unsigned char * RefDataPtr2) {
+  ogg_uint32_t Res; /* Final result */
+  unsigned long long int *src = (unsigned long long int*)SrcData;
+  unsigned long long int *ref1 = (unsigned long long int*)RefDataPtr1;
+  unsigned long long int *ref2 = (unsigned long long int*)RefDataPtr2;
+
+  __asm__("\n\t movq %1,%%mm0        \t # load src"
+	  "\n\t movq %2,%%mm1        \t # load ref1"
+	  "\n\t pavgb %3,%%mm1       \t # average of ref1 et ref2"
+	  "\n\t psadbw %%mm1,%%mm0   \t # does the sum of abs diff"
+	  "\n\t movd %%mm0, %0       \t # store result"
+	  : "=m" (Res) /* this is %0, output */
+	  : "m" (*src),      /* this is %1, src */
+	  "m" (*ref1),      /* this is %2, ref1 */
+	  "m" (*ref2)      /* this is %3, ref2 */
+	  );
+
+  return Res;
+}
+static inline void finish8_mmxext(void) {
+  /* Resets floating usage */
+  __asm__("emms" : : );
+}
+
+/*
+ * Extended MMX implementation wrapper and/or test functions
+ */
+
+/* Optional back to back testing of the C and MMX-ext implementation */
+static inline ogg_uint32_t sum_abs_diff8(unsigned char * NewDataPtr, 
+					 unsigned char * RefDataPtr) {
+  ogg_uint32_t res_mmxext;
+  res_mmxext = sum_abs_diff8_mmxext(NewDataPtr, RefDataPtr);
+#ifdef TEST_MMX
+  {
+    ogg_uint32_t res_c;
+    res_c = sum_abs_diff8_c(NewDataPtr, RefDataPtr);
+    if (res_c != res_mmxext)
+      fprintf(stderr,"sum_abs_diff8: res_c=%i, res_mmxext=%i\n",res_c,res_mmxext);
+  }
+#endif
+  return res_mmxext;
+}
+static inline ogg_uint32_t sum_abs_diffmean8(unsigned char * SrcData, 
+					     unsigned char * RefDataPtr1,
+					     unsigned char * RefDataPtr2) {
+  ogg_uint32_t res_mmxext;
+  res_mmxext = sum_abs_diffmean8_mmxext(SrcData, RefDataPtr1, RefDataPtr2);
+#ifdef TEST_MMX
+  {
+    ogg_uint32_t res_c;
+    res_c = sum_abs_diffmean8_c(SrcData, RefDataPtr1, RefDataPtr2);
+    if (abs(res_c - res_mmxext) > 8)
+      fprintf(stderr,"\nsum_abs_diffmean8: res_c=%i, res_mmxext=%i\n",res_c,res_mmxext);
+  }
+#endif
+  return res_mmxext;
+}
+static inline void finish8(void) {
+#ifdef TEST_MMX
+  finish8_c();
+#endif
+  finish8_mmxext();
+}
+
+#endif /* HAVE_MMXEXT */
+
+/*
+ * MMX implementation
+ */
+
+static unsigned long long int no_high_bits = 0x7F7F7F7F7F7F7F7FULL;
+
+/* MMX assembler implementation of the operation */
+static inline ogg_uint32_t sum_abs_diff8_mmx(unsigned char * NewDataPtr, 
+					     unsigned char  * RefDataPtr) {
+  unsigned long long int *new = (unsigned long long int*)NewDataPtr;
+  unsigned long long int *ref = (unsigned long long int*)RefDataPtr;
+  unsigned long long int mmtemp_d; /* diff result */
+  unsigned char* mmtemp = (unsigned char*) & mmtemp_d; /* for easy access */
+  ogg_uint32_t Res; /* Final result */
+  __asm__("\n\t movq %1,%%mm0        \t # load new"
+	  "\n\t movq %2,%%mm1        \t # load ref"
+	  "\n\t movq %%mm0,%%mm2     \t # copy new data"
+	  "\n\t psubusb %%mm1,%%mm0  \t # new - ref"
+	  "\n\t psubusb %%mm2,%%mm1  \t # ref - new"
+	  "\n\t por %%mm1,%%mm0      \t # get abs() via or (aka +)"
+	  "\n\t movq %%mm0, %0       \t # store result"
+	  : "=m" (mmtemp_d) /* this is %0, output */
+	  : "m" (*new),      /* this is %1, new */
+	  "m" (*ref)      /* this is %2, ref */
+	  );
+  /* Accumulate the result */
+  Res = mmtemp[0]; /* TODO: Possible improvement here? */
+  Res += mmtemp[1];
+  Res += mmtemp[2];
+  Res += mmtemp[3];
+  Res += mmtemp[4];
+  Res += mmtemp[5];
+  Res += mmtemp[6];
+  Res += mmtemp[7];
+
+  return Res;
+}
+
+static inline ogg_uint32_t sum_abs_diffmean8_mmx(unsigned char * SrcData, 
+						 unsigned char * RefDataPtr1,
+						 unsigned char * RefDataPtr2) {
+  ogg_uint32_t Res; /* Final result */
+  unsigned long long int *src = (unsigned long long int*)SrcData;
+  unsigned long long int *ref1 = (unsigned long long int*)RefDataPtr1;
+  unsigned long long int *ref2 = (unsigned long long int*)RefDataPtr2;
+  unsigned long long int mmtemp_d; /* diff result */
+  unsigned char* mmtemp = (unsigned char*) & mmtemp_d; /* for easy access */
+
+  __asm__("\n\t movq %1,%%mm0        \t # load src"
+	  "\n\t movq %2,%%mm1        \t # load ref1"
+	  "\n\t movq %3,%%mm2        \t # load ref2"
+	  "\n\t movq %%mm0,%%mm3     \t # copy src data"
+	  /* We do the division before the addition to avoid overflow
+	     and this can induce a by 1 error (replicated 8 times on
+	     the 64bits quantity */
+          "\n\t psrlq $1,%%mm1       \t # ref1/2"
+	  "\n\t pand %4,%%mm1"
+          "\n\t psrlq $1,%%mm2       \t # ref2/2"
+	  "\n\t pand %4,%%mm2"
+	  "\n\t paddusb %%mm2,%%mm1  \t # ref1/2 + ref2/2 = sum/2"
+	  "\n\t psubusb %%mm1,%%mm0  \t # src - sum/2"
+	  "\n\t psubusb %%mm3,%%mm1  \t # sum/2 - src"
+	  "\n\t por %%mm1,%%mm0      \t # get abs() via or (ie: +)"
+	  "\n\t movq %%mm0, %0       \t # store result"
+	  : "=m" (mmtemp_d) /* this is %0, output */
+	  : "m" (*src),      /* this is %1, src */
+	  "m" (*ref1),      /* this is %2, ref1 */
+	  "m" (*ref2),      /* this is %3, ref2 */
+	  "m" (no_high_bits)  /* this is %4, constant */
+	  );
+  /* Accumulate the result */
+  Res = mmtemp[0];  /* TODO: Possible improvement here? */
+  Res += mmtemp[1];
+  Res += mmtemp[2];
+  Res += mmtemp[3];
+  Res += mmtemp[4];
+  Res += mmtemp[5];
+  Res += mmtemp[6];
+  Res += mmtemp[7];
+
+  return Res;
+}
+
+static inline void accsumsqr_diff8_mmx(unsigned char *newc,
+				       unsigned char *refc,
+				       ogg_int32_t *accsum, ogg_int32_t *accsum2) {
+  unsigned long long int *new = (unsigned long long int*)newc;
+  unsigned long long int *ref = (unsigned long long int*)refc;
+  unsigned long long int mmtemp1_d; /* diff result */
+  unsigned long long int mmtemp2_d; /* diff^2 result */
+  ogg_int16_t* mmtemp1 = (ogg_int16_t*) & mmtemp1_d;
+  ogg_int32_t* mmtemp2 = (ogg_int32_t*) & mmtemp2_d;
+  __asm__("\n\t movq %2,%%mm0        \t # load new"
+	  "\n\t movq %3,%%mm1        \t # load ref"
+	  "\n\t pxor %%mm4,%%mm4     \t # zeroes mm4"
+	  "\n\t movq %%mm0,%%mm2     \t # copy new"
+	  "\n\t movq %%mm1,%%mm3     \t # copy ref"
+	  "\n\t punpcklbw %%mm4,%%mm0 \t # new first half in mm0"
+	  "\n\t punpcklbw %%mm4,%%mm1 \t # ref first half in mm1"
+	  "\n\t punpckhbw %%mm4,%%mm2 \t # new second half in mm2"
+	  "\n\t punpckhbw %%mm4,%%mm3 \t # ref second half in mm3"
+	  "\n\t psubw %%mm1,%%mm0    \t # new - ref (1st half)"
+	  "\n\t psubw %%mm3,%%mm2    \t # new - ref (2nd half)"
+	  "\n\t movq %%mm0,%%mm5     \t # copy 1st half"
+	  "\n\t movq %%mm2,%%mm6     \t # copy 2nd half"
+	  "\n\t pmaddwd %%mm5,%%mm5  \t # square+sum 1st half"
+	  "\n\t pmaddwd %%mm6,%%mm6  \t # square+sum 2nd half"
+	  "\n\t paddw %%mm2,%%mm0    \t # sum halves"
+	  "\n\t paddd %%mm6,%%mm5    \t # sum halves"
+	  "\n\t movq %%mm0, %0       \t # store 1st result"
+	  "\n\t movq %%mm5, %1       \t # store 2nd result"
+	  : "=m" (mmtemp1_d), /* this is %0, output */
+	  "=m" (mmtemp2_d)    /* this is %1, output */
+	  : "m" (*new),       /* this is %2, new */
+	  "m" (*ref)          /* this is %3, ref */
+	  );
+  /* Accumulate the result in the target accumulators */
+  {
+    ogg_int32_t t,t2;
+    t = 0; t2 = 0;
+    t += mmtemp1[0];
+    t += mmtemp1[1];
+    t += mmtemp1[2];
+    t += mmtemp1[3];
+    t2 += mmtemp2[0];
+    t2 += mmtemp2[1];
+    (*accsum) += t;
+    (*accsum2) += t2;
+  }
+}
+
+static inline void accsumsqr_diffmean8_mmx(unsigned char *newc,
+					   unsigned char *ref1c,
+					   unsigned char *ref2c,
+					   ogg_int32_t *accsum,
+					   ogg_int32_t *accsum2) {
+  unsigned long long int *new = (unsigned long long int*)newc;
+  unsigned long long int *ref1 = (unsigned long long int*)ref1c;
+  unsigned long long int *ref2 = (unsigned long long int*)ref2c;
+  unsigned long long int mmtemp1_d; /* diff result */
+  unsigned long long int mmtemp2_d; /* diff^2 result */
+  ogg_int16_t* mmtemp1 = (ogg_int16_t*) & mmtemp1_d;
+  ogg_int32_t* mmtemp2 = (ogg_int32_t*) & mmtemp2_d;
+  __asm__("\n\t movq %2,%%mm0        \t # load new"
+	  "\n\t movq %3,%%mm1        \t # load ref1"
+	  "\n\t movq %4,%%mm5        \t # load ref2"
+	  /* We do the division before the addition to avoid overflow
+	     and this can induce a by 1 error (replicated 8 times on
+	     the 64bits quantity */
+          "\n\t psrlq $1,%%mm1       \t # ref1/2"
+	  "\n\t pand %5,%%mm1"
+          "\n\t psrlq $1,%%mm5       \t # ref2/2"
+	  "\n\t pand %5,%%mm5"
+	  "\n\t paddusb %%mm5,%%mm1  \t # ref1/2 + ref2/2 = sum/2"
+	  "\n\t pxor %%mm4,%%mm4     \t # zeroes mm4"
+	  "\n\t movq %%mm0,%%mm2     \t # copy new"
+	  "\n\t movq %%mm1,%%mm3     \t # copy sum/2 aka ref"
+	  "\n\t punpcklbw %%mm4,%%mm0 \t # new first half in mm0"
+	  "\n\t punpcklbw %%mm4,%%mm1 \t # ref first half in mm1"
+	  "\n\t punpckhbw %%mm4,%%mm2 \t # new second half in mm2"
+	  "\n\t punpckhbw %%mm4,%%mm3 \t # ref second half in mm3"
+	  "\n\t psubw %%mm1,%%mm0    \t # new - ref (1st half)"
+	  "\n\t psubw %%mm3,%%mm2    \t # new - ref (2nd half)"
+	  "\n\t movq %%mm0,%%mm5     \t # copy 1st half"
+	  "\n\t movq %%mm2,%%mm6     \t # copy 2nd half"
+	  "\n\t pmaddwd %%mm5,%%mm5  \t # square+sum 1st half"
+	  "\n\t pmaddwd %%mm6,%%mm6  \t # square+sum 2nd half"
+	  "\n\t paddw %%mm2,%%mm0    \t # sum halves"
+	  "\n\t paddd %%mm6,%%mm5    \t # sum halves"
+	  "\n\t movq %%mm0, %0       \t # store result"
+	  "\n\t movq %%mm5, %1       \t # store result"
+	  : "=m" (mmtemp1_d), /* this is %0, output */
+	  "=m" (mmtemp2_d)    /* this is %1, output */
+	  : "m" (*new),       /* this is %2, new */
+	  "m" (*ref1),        /* this is %3, ref1 */
+	  "m" (*ref2),        /* this is %4, ref2 */
+	  "m" (no_high_bits)  /* this is %5, constant */
+	  );
+  /* Accumulate the result in the target accumulators */
+  {
+    ogg_int32_t t,t2;
+    t = 0; t2 = 0;
+    t += mmtemp1[0];
+    t += mmtemp1[1];
+    t += mmtemp1[2];
+    t += mmtemp1[3];
+    t2 += mmtemp2[0];
+    t2 += mmtemp2[1];
+    (*accsum) += t;
+    (*accsum2) += t2;
+  }
+}
+
+static inline void accsumsqr8_mmx(unsigned char *diffc,
+				  ogg_int32_t *accsum, ogg_int32_t *accsum2) {
+  unsigned long long int *diff = (unsigned long long int*)diffc;
+  unsigned long long int mmtemp1_d; /* sum result */
+  unsigned long long int mmtemp2_d; /* ^2 sum result */
+  ogg_int16_t* mmtemp1 = (ogg_int16_t*) & mmtemp1_d;
+  ogg_int32_t* mmtemp2 = (ogg_int32_t*) & mmtemp2_d;
+  __asm__("\n\t movq %2,%%mm0        \t # load diff"
+	  "\n\t pxor %%mm4,%%mm4     \t # zeroes mm4"
+	  "\n\t movq %%mm0,%%mm2     \t # copy diff"
+	  "\n\t punpcklbw %%mm4,%%mm0 \t # diff first half in mm0"
+	  "\n\t punpckhbw %%mm4,%%mm2 \t # diff second half in mm2"
+	  "\n\t movq %%mm0,%%mm5     \t # copy 1st half"
+	  "\n\t movq %%mm2,%%mm6     \t # copy 2nd half"
+	  "\n\t pmaddwd %%mm5,%%mm5   \t # square+sum 1st half"
+	  "\n\t pmaddwd %%mm6,%%mm6   \t # square+sum 2nd half"
+	  "\n\t paddw %%mm2,%%mm0    \t # sum base halves"
+	  "\n\t paddd %%mm6,%%mm5    \t # sum square halves"
+	  "\n\t movq %%mm0, %0       \t # store result"
+	  "\n\t movq %%mm5, %1       \t # store result"
+	  : "=m" (mmtemp1_d), /* this is %0, output */
+	  "=m" (mmtemp2_d)    /* this is %1, output */
+	  : "m" (*diff)       /* this is %2, diff */
+	  );
+  /* Accumulate the result in the target accumulators */
+  {
+    ogg_int32_t t,t2;
+    t = 0; t2 = 0;
+    t += mmtemp1[0];
+    t += mmtemp1[1];
+    t += mmtemp1[2];
+    t += mmtemp1[3];
+    t2 += mmtemp2[0];
+    t2 += mmtemp2[1];
+    (*accsum) += t;
+    (*accsum2) += t2;
+  }
+}
+
+static inline void finish8_mmx(void) {
+  /* Resets floating usage */
+  __asm__("emms" : : );
+}
+
+#ifndef HAVE_MMXEXT
+/*
+ * MMX implementation wrapper and or test functions
+ */
+
+/* Optional back to back testing of the C and MMX implementation */
+static inline ogg_uint32_t sum_abs_diff8(unsigned char * NewDataPtr, 
+					 unsigned char * RefDataPtr) {
+  ogg_uint32_t res_mmx;
+  res_mmx = sum_abs_diff8_mmx(NewDataPtr, RefDataPtr);
+#ifdef TEST_MMX
+  {
+    ogg_uint32_t res_c;
+    res_c = sum_abs_diff8_c(NewDataPtr, RefDataPtr);
+    if (res_c != res_mmx)
+      fprintf(stderr,"sum_abs_diff8: res_c=%i, res_mmx=%i\n",res_c,res_mmx);
+  }
+#endif
+  return res_mmx;
+}
+static inline ogg_uint32_t sum_abs_diffmean8(unsigned char * SrcData, 
+					     unsigned char * RefDataPtr1,
+					     unsigned char * RefDataPtr2) {
+  ogg_uint32_t res_mmx;
+  res_mmx = sum_abs_diffmean8_mmx(SrcData, RefDataPtr1, RefDataPtr2);
+#ifdef TEST_MMX
+  {
+    ogg_uint32_t res_c;
+    res_c = sum_abs_diffmean8_c(SrcData, RefDataPtr1, RefDataPtr2);
+    if (abs(res_c - res_mmx) > (1*8))
+      fprintf(stderr,"sum_abs_diffmean8: res_c=%i, res_mmx=%i\n",res_c,res_mmx);
+  }
+#endif
+  return res_mmx;
+}
+static inline void finish8(void) {
+#ifdef TEST_MMX
+  finish8_c();
+#endif
+  finish8_mmx();
+}
+
+#endif /* not HAVE_MMXEXT */
+
+/* MMX-based function that are always used (no improvement
+   with MMXEXT yet or ever) */
+static inline void accsumsqr_diff8(unsigned char *newc,
+				   unsigned char *refc,
+				   ogg_int32_t *accsum, ogg_int32_t *accsum2) {
+#ifdef TEST_MMX
+  ogg_int32_t accsum_c = (*accsum);
+  ogg_int32_t accsum2_c = (*accsum2);
+#endif
+  accsumsqr_diff8_mmx(newc,refc,accsum,accsum2);
+#ifdef TEST_MMX
+  {
+    accsumsqr_diff8_c(newc,refc,&accsum_c,&accsum2_c);
+    if ((abs(accsum_c - (*accsum)) > 0)||(abs(accsum2_c - (*accsum2)) > 0))
+      fprintf(stderr,"\nsaccsumsqr_diff8: accsum_c=%i, accsum_mmx=%i, "
+	      "accsum2_c=%i, accsum2_mmx=%i\n",
+	      accsum_c,(*accsum),accsum2_c,(*accsum2));
+  }
+#endif
+}
+
+static inline void accsumsqr_diffmean8(unsigned char *new,
+				       unsigned char *ref1,
+				       unsigned char *ref2,
+				       ogg_int32_t *accsum,
+				       ogg_int32_t *accsum2) {
+#ifdef TEST_MMX
+  ogg_int32_t accsum_c = (*accsum);
+  ogg_int32_t accsum2_c = (*accsum2);
+#endif
+  accsumsqr_diffmean8_mmx(new,ref1,ref2,accsum,accsum2);
+#ifdef TEST_MMX
+  {
+    accsumsqr_diffmean8_c(new,ref1,ref2,&accsum_c,&accsum2_c);
+    if ((abs(accsum_c - (*accsum)) > 8)||
+	/* empirical value ( 8*((2^16-1)^2 - (2^16-2)^2) ???) */
+	(abs(accsum2_c - (*accsum2)) > (5000)))
+      fprintf(stderr,"\nsaccsumsqr_diffmean8: accsum_c=%i, accsum_mmx=%i, "
+	      "accsum2_c=%i, accsum2_mmx=%i\n",
+	      accsum_c,(*accsum),accsum2_c,(*accsum2));
+  }
+#endif
+}
+
+static inline void accsumsqr8(unsigned char *diffc,
+			      ogg_int32_t *accsum, ogg_int32_t *accsum2) {
+  /* NB: Wrt performance, the impact of this function does not seem very
+   * interesting (very limited improvement).
+   */
+#ifdef TEST_MMX
+  ogg_int32_t accsum_c = (*accsum);
+  ogg_int32_t accsum2_c = (*accsum2);
+#endif
+  accsumsqr8_mmx(diffc,accsum,accsum2);
+#ifdef TEST_MMX
+  {
+    accsumsqr8_c(diffc,&accsum_c,&accsum2_c);
+    if ((abs(accsum_c - (*accsum)) > 0)||(abs(accsum2_c - (*accsum2)) > 0))
+      fprintf(stderr,"\nsaccsumsqr8: accsum_c=%i, accsum_mmx=%i, "
+	      "accsum2_c=%i, accsum2_mmx=%i\n",
+	      accsum_c,(*accsum),accsum2_c,(*accsum2));
+  }
+#endif
+}
+
+#else /* HAVE_MMX */
+
+#ifdef HAVE_MMXEXT
+#error HAVE_MMXEXT defined while HAVE_MMX undefined!
+#endif
+
+/*
+ * C wrapper functions
+ */
+static inline ogg_uint32_t sum_abs_diff8(unsigned char * NewDataPtr, 
+					 unsigned char  * RefDataPtr) {
+  return sum_abs_diff8_c(NewDataPtr, RefDataPtr);
+}
+static inline ogg_uint32_t sum_abs_diffmean8(unsigned char * SrcData, 
+					     unsigned char * RefDataPtr1,
+					     unsigned char * RefDataPtr2) {
+  return sum_abs_diffmean8_c(SrcData, RefDataPtr1, RefDataPtr2);
+}
+static inline void accsumsqr_diff8(unsigned char *new, unsigned char *ref,
+				   ogg_int32_t *accsum, ogg_int32_t *accsum2) {
+  accsumsqr_diff8_c(new,ref,accsum,accsum2);
+}
+static inline void accsumsqr_diffmean8(unsigned char *new,
+				   unsigned char *ref1, unsigned char *ref2,
+				   ogg_int32_t *accsum, ogg_int32_t *accsum2) {
+  accsumsqr_diffmean8_c(new,ref1,ref2,accsum,accsum2);
+}
+static inline void accsumsqr8(unsigned char *diff,
+			      ogg_int32_t *accsum, ogg_int32_t *accsum2) {
+  /* NB: Wrt performance, the impact of this function does not seem very
+   * interesting (very limited improvement).
+   */
+  accsumsqr8_c(diff,accsum,accsum2);
+}
+static inline void finish8(void) {
+  finish8_c();
+}
+
+#endif /* HAVE_MMX */
+
+
 /* Initialises motion compentsation. */
 void InitMotionCompensation ( CP_INSTANCE *cpi ){
   int i;
@@ -110,100 +786,36 @@
      second reference pointer */
   if ( AbsRefOffset == 0 ) {
     for ( i=0; i<BLOCK_HEIGHT_WIDTH; i++ ) {
-      DiffVal = ((int)NewDataPtr[0]) - (int)RefDataPtr1[0];
-      XSum += DiffVal;
 
-      /* negative array indexes are strictly forbidden by ANSI C and C99 */
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[1]) - (int)RefDataPtr1[1];
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[2]) - (int)RefDataPtr1[2];
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[3]) - (int)RefDataPtr1[3];
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[4]) - (int)RefDataPtr1[4];
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[5]) - (int)RefDataPtr1[5];
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[6]) - (int)RefDataPtr1[6];
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[7]) - (int)RefDataPtr1[7];
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
+      accsumsqr_diff8(NewDataPtr,RefDataPtr1,&XSum,&XXSum);
+
       /* Step to next row of block. */
       NewDataPtr += PixelsPerLine;
       RefDataPtr1 += STRIDE_EXTRA + PixelsPerLine;
     }
+    finish8();
 
   }else{
 
     /* Simple two reference interpolation */
     for ( i=0; i<BLOCK_HEIGHT_WIDTH; i++ ) {
-      DiffVal = ((int)NewDataPtr[0]) - 
-	(((int)RefDataPtr1[0] + (int)RefDataPtr2[0]) / 2);
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[1]) - 
-	(((int)RefDataPtr1[1] + (int)RefDataPtr2[1]) / 2);
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[2]) - 
-	(((int)RefDataPtr1[2] + (int)RefDataPtr2[2]) / 2);
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[3]) - 
-	(((int)RefDataPtr1[3] + (int)RefDataPtr2[3]) / 2);
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-
-      DiffVal = ((int)NewDataPtr[4]) - 
-	(((int)RefDataPtr1[4] + (int)RefDataPtr2[4]) / 2);
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[5]) - 
-	(((int)RefDataPtr1[5] + (int)RefDataPtr2[5]) / 2);
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[6]) - 
-	(((int)RefDataPtr1[6] + (int)RefDataPtr2[6]) / 2);
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
-      DiffVal = ((int)NewDataPtr[7]) -
-	(((int)RefDataPtr1[7] + (int)RefDataPtr2[7]) / 2);
-      XSum += DiffVal;
-      XXSum += DiffVal*DiffVal;
-      
+
+      accsumsqr_diffmean8(NewDataPtr,RefDataPtr1,RefDataPtr2,
+			  &XSum,&XXSum);
+
       /* Step to next row of block. */
       NewDataPtr += PixelsPerLine;
       RefDataPtr1 += STRIDE_EXTRA+PixelsPerLine;
       RefDataPtr2 += STRIDE_EXTRA+PixelsPerLine;
     }
+    finish8();
   }
   
   /* Compute and return population variance as mis-match metric. */
   return (( (XXSum<<6) - XSum*XSum ));
 }
 
+
 static ogg_uint32_t GetSumAbsDiffs  (unsigned char * NewDataPtr, 
 			      unsigned char  * RefDataPtr, 
 			      ogg_uint32_t PixelsPerLine, 
@@ -213,20 +825,15 @@
   
   /* Decide on standard or MMX implementation */
   for ( i=0; i < BLOCK_HEIGHT_WIDTH; i++ ) {
-    DiffVal += abs( ((int)NewDataPtr[0]) - ((int)RefDataPtr[0]) );
-    DiffVal += abs( ((int)NewDataPtr[1]) - ((int)RefDataPtr[1]) );
-    DiffVal += abs( ((int)NewDataPtr[2]) - ((int)RefDataPtr[2]) );
-    DiffVal += abs( ((int)NewDataPtr[3]) - ((int)RefDataPtr[3]) );
-    DiffVal += abs( ((int)NewDataPtr[4]) - ((int)RefDataPtr[4]) );
-    DiffVal += abs( ((int)NewDataPtr[5]) - ((int)RefDataPtr[5]) );
-    DiffVal += abs( ((int)NewDataPtr[6]) - ((int)RefDataPtr[6]) );
-    DiffVal += abs( ((int)NewDataPtr[7]) - ((int)RefDataPtr[7]) );
+    DiffVal += sum_abs_diff8(NewDataPtr,RefDataPtr);
 
     /* Step to next row of block. */
     NewDataPtr += PixelsPerLine;
     RefDataPtr += STRIDE_EXTRA+PixelsPerLine;
   }
 
+  finish8();
+
   return DiffVal;  
 }
 
@@ -239,14 +846,7 @@
   ogg_uint32_t	DiffVal = ErrorSoFar;
   
   for ( i=0; i < BLOCK_HEIGHT_WIDTH; i++ ) {
-    DiffVal += abs( ((int)NewDataPtr[0]) - ((int)RefDataPtr[0]) );
-    DiffVal += abs( ((int)NewDataPtr[1]) - ((int)RefDataPtr[1]) );
-    DiffVal += abs( ((int)NewDataPtr[2]) - ((int)RefDataPtr[2]) );
-    DiffVal += abs( ((int)NewDataPtr[3]) - ((int)RefDataPtr[3]) );
-    DiffVal += abs( ((int)NewDataPtr[4]) - ((int)RefDataPtr[4]) );
-    DiffVal += abs( ((int)NewDataPtr[5]) - ((int)RefDataPtr[5]) );
-    DiffVal += abs( ((int)NewDataPtr[6]) - ((int)RefDataPtr[6]) );
-    DiffVal += abs( ((int)NewDataPtr[7]) - ((int)RefDataPtr[7]) );
+    DiffVal += sum_abs_diff8(NewDataPtr,RefDataPtr);
 
     if ( DiffVal > BestSoFar )break;
 
@@ -255,6 +855,8 @@
     RefDataPtr += STRIDE_EXTRA+PixelsPerLine;
   }
 
+  finish8();
+
   return DiffVal;  
 }
 
@@ -276,22 +878,8 @@
 			       ErrorSoFar);
   } else  {
     for ( i=0; i < BLOCK_HEIGHT_WIDTH; i++ ) {
-      DiffVal += abs( ((int)SrcData[0]) - (((int)RefDataPtr1[0] + 
-					    (int)RefDataPtr2[0]) / 2) );
-      DiffVal += abs( ((int)SrcData[1]) - (((int)RefDataPtr1[1] + 
-					    (int)RefDataPtr2[1]) / 2) );
-      DiffVal += abs( ((int)SrcData[2]) - (((int)RefDataPtr1[2] + 
-					    (int)RefDataPtr2[2]) / 2) );
-      DiffVal += abs( ((int)SrcData[3]) - (((int)RefDataPtr1[3] + 
-					    (int)RefDataPtr2[3]) / 2) );
-      DiffVal += abs( ((int)SrcData[4]) - (((int)RefDataPtr1[4] + 
-					    (int)RefDataPtr2[4]) / 2) );
-      DiffVal += abs( ((int)SrcData[5]) - (((int)RefDataPtr1[5] + 
-					    (int)RefDataPtr2[5]) / 2) );
-      DiffVal += abs( ((int)SrcData[6]) - (((int)RefDataPtr1[6] + 
-					    (int)RefDataPtr2[6]) / 2) );
-      DiffVal += abs( ((int)SrcData[7]) - (((int)RefDataPtr1[7] + 
-					    (int)RefDataPtr2[7]) / 2) );
+
+      DiffVal += sum_abs_diffmean8(SrcData,RefDataPtr1,RefDataPtr2);
       
       if ( DiffVal > BestSoFar ) break;
 
@@ -302,6 +890,8 @@
     }
   }
 
+  finish8();
+
   return DiffVal;  
 }
 
@@ -316,28 +906,14 @@
   DiffPtr = DataPtr;
   
   for ( i=0; i<BLOCK_HEIGHT_WIDTH; i++ ) {
-	
-    /* Examine alternate pixel locations. */
-    XSum += DiffPtr[0];
-    XXSum += DiffPtr[0]*DiffPtr[0];
-    XSum += DiffPtr[1];
-    XXSum += DiffPtr[1]*DiffPtr[1];
-    XSum += DiffPtr[2];
-    XXSum += DiffPtr[2]*DiffPtr[2];
-    XSum += DiffPtr[3];
-    XXSum += DiffPtr[3]*DiffPtr[3];
-    XSum += DiffPtr[4];
-    XXSum += DiffPtr[4]*DiffPtr[4];
-    XSum += DiffPtr[5];
-    XXSum += DiffPtr[5]*DiffPtr[5];
-    XSum += DiffPtr[6];
-    XXSum += DiffPtr[6]*DiffPtr[6];
-    XSum += DiffPtr[7];
-    XXSum += DiffPtr[7]*DiffPtr[7];
+
+    accsumsqr8(DiffPtr,&XSum,&XXSum);
 
     /* Step to next row of block. */
     DiffPtr += PixelsPerLine;
   }
+
+  finish8();
 
   /* Compute population variance as mis-match metric. */
   return (( (XXSum<<6) - XSum*XSum ) );
